local AddonName, Hermes = ...
local L = LibStub('AceLocale-3.0'):GetLocale('Hermes')

function Hermes:StopTrackingAllAbilities()
	while #Hermes.Abilities > 0 do
		self:StopTrackingAbility(Hermes.Abilities[1].dbp)
	end
end

function Hermes:FindTrackedAbility(dbp)
	for _, ability in ipairs(Hermes.Abilities) do
		if ability.dbp == dbp then
			return ability
		end
	end
	
	return nil
end

function Hermes:StartTrackingAbility(dbp)
	if not dbp then error("null ability") end
	
	ability = {
		id = dbp.id,
		name = dbp.name,
		class = dbp.class,
		icon = dbp.icon,
		dbp = dbp,
	}
									
	tinsert(Hermes.Abilities, ability)
	
	--call events
	Hermes:FireEvent("OnAbilityAdded", ability)
	
	--if receiving, then request update from senders
	if(Hermes.Receiving == true) then
		Hermes:RequestAbilityUpdate(ability.id)
	end
end

function Hermes:GetAbilityStats(ability)
	local min_available_time = nil
	local instances_total = 0
	local instances_oncooldown = 0
	local instances_available = 0
	local instances_unavailable = 0
	local is_available = 0
	
	for _, instance in ipairs(Hermes.AbilityInstances) do
		if instance.ability == ability then
			local senderAvailable = self:IsSenderAvailable(instance.sender) --does not take ability being on cooldown into account, just whether they're alive, online, and in same zone
			
			--set instances_total
			instances_total = instances_total + 1
			
			--instances_unavailable
			if not senderAvailable then
				instances_unavailable = instances_unavailable + 1
			end
			
			--update min_available_time, but only if the sender is available and there is a remaining time
			if senderAvailable and instance.remaining then
				if not min_available_time then
					min_available_time = instance.remaining
				else
					if instance.remaining < min_available_time then
						min_available_time = instance.remaining
					end
				end
			end
			
			if senderAvailable and not instance.remaining then
				instances_available = instances_available + 1
			end
			
			if senderAvailable and instance.remaining then
				instances_oncooldown = instances_oncooldown + 1
			end
			
		end
	end

	return min_available_time, instances_total, instances_oncooldown, instances_available, instances_unavailable
end

function Hermes:StopTrackingAbility(dbp)
	if not dbp then error("null dbp") end
	
	local ability = self:FindTrackedAbility(dbp)
	
	if not ability then error("null ability") end
	
	--first remove any ability instances
	self:RemoveAbilityInstancesForAbility(ability)
	
	--call events
	Hermes:FireEvent("OnAbilityRemoved", ability)
	
	--now delete from table
	Hermes:DeleteFromIndexedTable(Hermes.Abilities, ability)
end

function Hermes:AddSpell(newSpellId, newSpellName, newSpellClass)
	--see if the spell exists
	local name
	local icon
	if(newSpellId) then
		name, _, icon, _, _, _, _, _, _ = GetSpellInfo(newSpellId)
	else
		--try to find the spell id for the given spell name
		local matches = self:GetSpellID(newSpellName)
		
		if(matches and #matches > 0) then
			--still need the icon
			newSpellId = matches[1]
			name, _, icon, _, _, _, _, _, _ = GetSpellInfo(newSpellId)
			if(#matches > 1) then
				--there are more than one spellid from the name given, ask the user which they want to use
				local message = L["|cFFFF0000Hermes Warning|r"].." "..L["multiple id's were found. The first id was chosen"]
				print(message)
			end
		end
	end
	
	if(name and icon) then
		--make sure we're not already tracking this spell
		local id = self:EncodeTrackerIdentifier(newSpellId, "spell")
		for i, spell in ipairs(self.db.profile.spells) do
			if(spell.id == id) then
				local message = L["|cFFFF0000Hermes Warning|r"].." "..L["spell has already been added"]
				print(message)
				return nil
			end
		end
		
		local spell = {
			class = newSpellClass,
			id = tonumber(id),
			name = name,
			icon = icon,
			enabled = true, --assume they want to track right away
		}
		
		--add to db
		tinsert(self.db.profile.spells, spell)
		
		--sort the spells
		table.sort(self.db.profile.spells, function(a, b) return Hermes:SortProfileSpells(a, b) end)

		--enable tracking
		self:StartTrackingAbility(spell)
		
		return 1
	else
		local message = L["|cFFFF0000Hermes Warning|r"].." "..L["spell name or id not found"]
		print(message)
		return nil
	end
end

function Hermes:DeleteSpell(dbp)
	--stop tracking it if needed
	local ability = self:FindTrackedAbility(dbp)
	if ability then
		self:StopTrackingAbility(dbp)
	end
	
	--delete from profile
	Hermes:DeleteFromIndexedTable(self.db.profile.spells, dbp)
end

function Hermes:AddItem(newItemId, newItemName, newItemClass)
	local id, name, icon = self:GetItemInfo(newItemId, newItemName)
	if(not icon) then
		local message = L["|cFFFF0000Hermes Warning|r"].." "..L["Item name or id not found. If you're confident the id or name is correct, try having someone link the item or putting the item in your inventory and adding again."]
		print(message)
		return nil
	end
	
	if(icon and id) then
		id = self:EncodeTrackerIdentifier(tonumber(id), "item")
		--make sure we're not already tracking this item
		for i, item in ipairs(self.db.profile.items) do
			if(item.id == tonumber(id)) then
				local message = L["|cFFFF0000Hermes Warning|r"].." "..L["Item has already been added"]
				print(message)
				return nil
			end
		end
		
		--we need to cache the item
		local caching = nil
		if(not name) then
			caching = 1
			name  = "..."..L["searching"].." ("..tostring(self:DecodeTrackerIdentifier(tonumber(id)))..")"
		end
		
		local item = {
			class = newItemClass,
			id = tonumber(id),
			name = name,
			icon = icon,
			enabled = true,
			caching = caching
		}
		
		--add to db
		tinsert(self.db.profile.items, item)
		
		--sort the spells
		table.sort(self.db.profile.items, function(a, b) return Hermes:SortProfileItems(a, b) end)

		--enable tracking
		self:StartTrackingAbility(item)
		
		--fire off the name cache timer if needed
		if(caching) then
			Hermes:StartItemNameTimer(item.id)
		end
		
		return 1
	end
	
	return nil
end

function Hermes:DeleteItem(dbp)
	--stop tracking it if needed
	local ability = self:FindTrackedAbility(dbp)
	if ability then
		self:StopTrackingAbility(dbp)
	end
	
	--delete from profile
	Hermes:DeleteFromIndexedTable(self.db.profile.items, dbp)
end
