local AddonName, Hermes = ...

local SCAN_FREQUENCY = 0.05
local DELTA_THRESHOLD_FOR_REMAINING_CHANGE = 3

local _keepRunning = false
local function OnAbilityInstanceFrameUpdate(frame, elapsed)
	frame.LastScan = frame.LastScan + elapsed;
	while (frame.LastScan > SCAN_FREQUENCY) do
		_keepRunning = false
		frame.LastScan = frame.LastScan - SCAN_FREQUENCY
		for _, instance in ipairs(Hermes.AbilityInstances) do
			if not instance.remaining then
				--do nothing, it's not on cooldown
			else
				--update remaining time
				instance.remaining = instance.initialDuration - (GetTime()- instance.initialTimeStamp)
			
				if instance.remaining <= 0 then
					--item just went off cooldown
					instance.remaining = nil
					instance.initialDuration = nil
					instance.initialTimeStamp = nil
					
					Hermes:FireEvent("OnAbilityAvailableSendersChanged", instance.ability) -- ADDED
					Hermes:FireEvent("OnAbilityInstanceStopCooldown", instance)
				else
					--item still on cooldown
					Hermes:FireEvent("OnAbilityInstanceUpdateCooldown", instance)
					
					--allow the OnUpdate script to continue running
					_keepRunning = true
				end
			end
		end
		
		--if there's nothing with a remaining time, then kill the script
		if _keepRunning == false then
			Hermes:StopTrackingAbilityInstances()
		end
	end
end

function Hermes:InitializeAbilityInstanceFrame()
	--this frame manages all OnUpdate callbacks
	Hermes.AbilityInstanceFrame = CreateFrame("Frame", nil, UIParent)
	Hermes.AbilityInstanceFrame:ClearAllPoints()
	Hermes.AbilityInstanceFrame:EnableMouse(false)
	Hermes.AbilityInstanceFrame:SetMovable(false)
	Hermes.AbilityInstanceFrame:SetToplevel(false)
end

function Hermes:StartTrackingAbilityInstances()
	Hermes.AbilityInstanceFrame.LastScan = 0
	Hermes.AbilityInstanceFrame:SetScript("OnUpdate", OnAbilityInstanceFrameUpdate)
end

function Hermes:StopTrackingAbilityInstances()
	Hermes.AbilityInstanceFrame:SetScript("OnUpdate", nil)
end

function Hermes:RemoveAllAbilityInstances()
	while #Hermes.AbilityInstances > 0 do
		self:RemoveAbilityInstances(Hermes.AbilityInstances[1])
	end
end

function Hermes:SetAbilityInstance(ability, sender, duration)
	--first see if one already exists
	local instance = self:FindAbilityInstance(ability, sender)
	local timeStamp = nil
	
	--only set a timestamp if there is an actual duration
	if duration then
		timeStamp = GetTime()
	end
	
	if not instance then
		--new instance, set it as though it's available initially, if it's not, another message will be sent with appropriate data
		instance = {
			ability = ability,
			sender = sender,
			initialDuration = nil,
			initialTimeStamp = nil,
			remaining = nil,
		}
		
		tinsert(Hermes.AbilityInstances, instance)

		--fire off an added event since it's new
		Hermes:FireEvent("OnAbilityTotalSendersChanged", instance.ability) -- ADDED
		Hermes:FireEvent("OnAbilityAvailableSendersChanged", instance.ability) -- ADDED
		Hermes:FireEvent("OnAbilityInstanceAdded", instance)
		
		--if it's on cooldown, then start it
		if duration then
			instance.initialDuration = duration
			instance.initialTimeStamp = timeStamp
			instance.remaining = duration
			Hermes:FireEvent("OnAbilityAvailableSendersChanged", instance.ability) -- ADDED
			Hermes:FireEvent("OnAbilityInstanceStartCooldown", instance)
		end
	else
		--existing instance, update it to reflect changes
		local lastRemaining = instance.remaining
		
		instance.initialDuration = duration
		instance.initialTimeStamp = timeStamp
		instance.remaining = duration

		--if it's on cooldown, then possibly start it
		if duration then
			--only start it if it wasn't on cooldown preivously, or if the difference in the previous cooldown value and the new one is greater than the specified threshold.
			--this helps cleanup events that can fire due to the Hermes handshaking that goes on
			if not lastRemaining or (lastRemaining and math.abs(lastRemaining - duration) > DELTA_THRESHOLD_FOR_REMAINING_CHANGE) then
				Hermes:FireEvent("OnAbilityAvailableSendersChanged", instance.ability) -- ADDED
				Hermes:FireEvent("OnAbilityInstanceStartCooldown", instance)
			end
		end
		
		--if it's not on cooldown, and it was on cooldown the last time we updated it, then fire off stop message.
		--otherwise there's no need to fire any events since status didn't change
		if not duration and lastRemaining then
			Hermes:FireEvent("OnAbilityAvailableSendersChanged", instance.ability) -- ADDED
			Hermes:FireEvent("OnAbilityInstanceStopCooldown", instance)
		end
	end
		
	--start OnUpdate script if not already running and this instance has a duration
	if duration and not Hermes.AbilityInstanceFrame:GetScript("OnUpdate") then
		self:StartTrackingAbilityInstances()
	end
end

function Hermes:FindAbilityInstance(ability, sender)
	for _, instance in ipairs(Hermes.AbilityInstances) do
		if(instance.ability == ability and instance.sender == sender) then
			return instance
		end
	end
	
	return nil
end

function Hermes:RemoveAbilityInstance(instance)
	--call the event
	Hermes:FireEvent("OnAbilityAvailableSendersChanged", instance.ability) -- ADDED
	Hermes:FireEvent("OnAbilityTotalSendersChanged", instance.ability) -- ADDED
	Hermes:FireEvent("OnAbilityInstanceRemoved", instance)
	
	--self:UpdateAbilityAvailabilityForAbility(instance.ability)
	
	--now delete from table
	Hermes:DeleteFromIndexedTable(Hermes.AbilityInstances, instance)
	
	--stop OnUpdate script if nothing to track
	if #Hermes.AbilityInstances == 0 and Hermes.AbilityInstanceFrame:GetScript("OnUpdate") then
		self:StopTrackingAbilityInstances()
	end
end

function Hermes:RemoveAbilityInstancesForAbility(ability)
	local items = {}
	
	--first find all the matches
	for _, instance in ipairs(Hermes.AbilityInstances) do
		if instance.ability == ability then
			tinsert(items, instance)
		end
	end
	
	--now remove each match
	for _, instance in ipairs(items) do
		self:RemoveAbilityInstance(instance)
	end
end

function Hermes:RemoveAbilityInstancesForSender(sender)
	local items = {}
	
	--first find all the matches
	for _, instance in ipairs(Hermes.AbilityInstances) do
		if instance.sender == sender then
			tinsert(items, instance)
		end
	end
	
	--now remove each match
	for _, instance in ipairs(items) do
		self:RemoveAbilityInstance(instance)
	end
end